
/* BEGIN COPYRIGHT NOTICE

Copyright (c) 1992 by John Sarraille and Peter DiFalco
(john@ishi.csustan.edu)

Copyright (c) 1998 by John Sarraille
(john@ishi.csustan.edu)

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that copyright notice and this permission
notice appear in supporting documentation.

The algorithm used in this program was inspired by the paper
entitled "A Fast Algorithm To Determine Fractal Dimensions By
Box Counting", which was written by Liebovitch and Toth, and
which appeared in the journal "Physics Letters A", volume 141,
pp 386-390, (1989).

This program is not warranted: use at your own risk.

END COPYRIGHT NOTICE */


/*

   FD3 PROGRAM FILE

Filename:  convert.c

OUTLINE OF CALLS:

GetIntFromString +
  IntCharsAreValid +
  ExtractValue +
GetULfromString +
  ULCharsAreValid +
  ExtractULvalue +
IsOptionString +

*/

#include "fd.h"

/*
     THE FUNCTIONS IN THIS FILE:
*/

int    GetIntFromString    () ;
int    IntCharsAreValid    () ;
int    ExtractIntValue     () ;
int    GetULfromString     () ;
int    ULCharsAreValid     () ;
int    ExtractULvalue      () ;
int    IsOptionString      () ;

/* ######################################################### */
/*                      GetIntFromString                     */
/* ######################################################### */
/*
       FUNCTION TO CONVERT A STRING TO INTEGER.
*/      
int GetIntFromString (stringPtr, maxInt, valuePtr)
       char  *  stringPtr ;
       int      maxInt ;
       int   *  valuePtr ;
{ 
  if (!IntCharsAreValid (stringPtr)) return(FALSE) ;
  else return (ExtractIntValue (stringPtr, maxInt, valuePtr));
}
/* ######################################################### */
/*                      IntCharsAreValid                     */
/* ######################################################### */
/*
       FUNCTION TO TEST A STRING TO SEE IF IT REPRESENTS AN
       INTEGER.  WE WANT A STRING WHICH IS AN OPTIONAL '+' OR
       '-', FOLLOWED BY A NON-EMPTY SEQUENCE OF DIGITS.
*/      
int  IntCharsAreValid (stringPtr)
       char  *  stringPtr ;
{ 
  int otherCharsOK, i, digitIsFirstChar, firstCharIsOK ;

  if (strlen(stringPtr)==0) return(FALSE) ;
  digitIsFirstChar = isdigit(stringPtr[0]) ;
  if (!digitIsFirstChar && (strlen(stringPtr)==1)) return(FALSE) ;
  firstCharIsOK = digitIsFirstChar || (stringPtr[0] == '+')
                                       || (stringPtr[0] == '-') ;
  otherCharsOK = TRUE ;
  for (i=1; i<strlen(stringPtr); i++)
    if ( ! isdigit(stringPtr[i]) ) otherCharsOK = FALSE ;

  return (firstCharIsOK && otherCharsOK) ;
}
/* ######################################################### */
/*                      ExtractIntValue                      */
/* ######################################################### */
/*
       FUNCTION TO COMPUTE THE INTEGER VALUE OF A STRING, USING
       HORNER'S ALGORITHM.  THE STRING IS ASSUMED TO HAVE NO
       ERRORS.
*/      
int ExtractIntValue (stringPtr, maxInt, valuePtr)
       char  *  stringPtr ;
       int      maxInt;
       int   *  valuePtr;
{ 
  int i, start, value;
  double testNum ;

  value = 0 ;
  if ( isdigit(stringPtr[0]) ) start = 0 ; else start = 1;
  for (i=start; i<strlen(stringPtr); i++)
  {
    testNum = maxInt-stringPtr[i] ;
    testNum = testNum + '0' ;
    testNum = testNum / 10 ;
    if (value > testNum) return(FALSE);
    value = (10 * value) - '0';
    value = value + stringPtr[i];
  }
  if (stringPtr[0] == '-') value = -value ;
  (* valuePtr) = value ;
  return(TRUE) ;
}
/* ######################################################### */
/*                      GetULfromString                      */
/* ######################################################### */
/*
       FUNCTION TO CONVERT A STRING TO AN UNSIGNED LONG INT
*/      
int GetULfromString (stringPtr, maxUlong, valuePtr)
       char  *  stringPtr ;
       uLong    maxUlong ;
       uLong *  valuePtr ;
{ 
  if (!ULCharsAreValid (stringPtr)) return(FALSE) ;
  else return (ExtractULvalue (stringPtr, maxUlong, valuePtr));
}
/* ######################################################### */
/*                       ULCharsAreValid                     */
/* ######################################################### */
/*
       FUNCTION TO TEST A STRING TO SEE IF IT REPRESENTS AN
       UNSIGNED LONG INT.  WE WANT A STRING WHICH IS AN
       OPTIONAL '+' FOLLOWED BY A NON-EMPTY SEQUENCE OF DIGITS.
*/      
int  ULCharsAreValid (stringPtr)
       char  *  stringPtr ;
{ 
  int otherCharsOK, i, digitIsFirstChar, firstCharIsOK ;

  if (strlen(stringPtr)==0) return(FALSE) ;
  digitIsFirstChar = isdigit(stringPtr[0]) ;
  if (!digitIsFirstChar && (strlen(stringPtr)==1)) return(FALSE) ;
  firstCharIsOK = digitIsFirstChar || (stringPtr[0] == '+') ;
  otherCharsOK = TRUE ;
  for (i=1; i<strlen(stringPtr); i++)
    if ( ! isdigit(stringPtr[i]) ) otherCharsOK = FALSE ;

  return (firstCharIsOK && otherCharsOK) ;
}
/* ######################################################### */
/*                       ExtractULvalue                      */
/* ######################################################### */
/*
       FUNCTION TO COMPUTE THE UNSIGNED LONG INT VALUE OF A
       STRING, USING HORNER'S ALGORITHM.  THE STRING IS ASSUMED
       TO HAVE NO ERRORS.
*/      
int ExtractULvalue (stringPtr, maxUlong, valuePtr)
       char   *  stringPtr ;
       uLong     maxUlong;
       uLong  *  valuePtr;
{ 
  int i, start ;
  uLong  value;
  double testNum ;

  value = 0 ;
  if ( isdigit(stringPtr[0]) ) start = 0 ; else start = 1;
  for (i=start; i<strlen(stringPtr); i++)
  {
    testNum = maxUlong-stringPtr[i] ;
    testNum = testNum + '0' ;
    testNum = testNum / 10 ;
    if (value > testNum) return(FALSE);
    value = (10 * value) - '0';
    value = value + stringPtr[i];
  }
  (* valuePtr) = value ;
  return(TRUE) ;
}
/* ######################################################### */
/*                       IsOptionString                      */
/* ######################################################### */
/*
       FUNCTION TO DETERMINE IF A STRING IS AN OPTION STRING
       (e.g. -aux).  IF THE STRING STARTS WITH '-' AND THE REST
       IS A SERIES OF ONE OR MORE ALPHA CHARACTERS, THEN WE
       CONCLUDE THAT IT IS AN OPTION STRING.
*/      
int IsOptionString   (ptrToString) 
   char * ptrToString ;
{
   int okSoFar, i ;

   okSoFar = (strlen(ptrToString) > 1) && (ptrToString[0] == '-') ;
   if (okSoFar)
   {
      i = 1 ;
      do
      {
        okSoFar = okSoFar && (isalpha(ptrToString[i])) ;
        i = i + 1 ;
      } while ((i<strlen(ptrToString)) && (okSoFar)) ;
   }
   return(okSoFar) ;
}
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */     
/*                        end convert.c                      */
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */
