
/* BEGIN COPYRIGHT NOTICE

Copyright (c) 1992 by John Sarraille and Peter DiFalco
(john@ishi.csustan.edu)

Copyright (c) 1998 by John Sarraille
(john@ishi.csustan.edu)

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that copyright notice and this permission
notice appear in supporting documentation.

The algorithm used in this program was inspired by the paper
entitled "A Fast Algorithm To Determine Fractal Dimensions By
Box Counting", which was written by Liebovitch and Toth, and
which appeared in the journal "Physics Letters A", volume 141,
pp 386-390, (1989).

This program is not warranted: use at your own risk.

END COPYRIGHT NOTICE */


/*
   FD3 PROGRAM FILE

Filename:  covers.c

OUTLINE OF CALLS:

InitCovData +
GetStatRAM +
  gmalloc
InitCellSizes +
PrintInitCovData +
SweepData +
  DBSweepData 
  StdDataSweep +
    FrontIndex
    InitStatRAM +
    SizeQ
    SEEKbitCHANGE  (macro)
    UPDATEcovSTATS  (macro)
    FinStats +
      FinCapStats +
      FinInfStats +
      FinCorrStats +
  mcovSweep 
FindRange +
  RangeSweep +
  ShowRange +
ReportCovData +
  CSTtitles +
  CSTbody +
    MakeStarBar +
  TwoPtEstTitles +
  TwoPtEstBody +
    MakeStarBar +
*/

#include "fd.h"
#include "covmacro.h"

/* 
    THE FUNCTIONS IN THIS FILE: 
*/
void  InitCovData     () ;
void  GetStatRAM      () ;
void  InitCellSizes   () ;
void  PrintInitCovData() ;
void  SweepData       () ;
void  StdDataSweep    () ;
void  InitStatRAM     () ;
void  FinStats        () ;
void  FinCapStats     () ;
void  FinInfStats     () ;
void  FinCorrStats    () ;
void  FindRange       () ;
void  RangeSweep      () ;
void  ShowRange       () ;
void  ReportCovData   () ;
void  CSTtitles       () ;
void  CSTbody         () ;
void  MakeStarBar     () ;
void  TwoPtEstTitles  () ;
void  TwoPtEstBody    () ;

/*   
      THE FUNCTIONS NOT IN THIS FILE,
      BUT CALLED FROM HERE:
*/
extern void   * gmalloc     () ;     /*  in mem.c      */
extern void     DBSweepData () ;     /*  in coversdb.c */
extern uLong    FrontIndex  () ;     /*  in queue.c    */
extern uLong    SizeQ       () ;     /*  in queue.c    */
extern void     mcovSweep   () ;     /*  in coversm.c  */
/*
       MACROS CALLED FROM THIS FILE

    SEEKbitCHANGE   (macro)          (*  covmacro.h   *)
    UPDATEcovSTATS  (macro)          (*  covmacro.h   *)
    ISa  (macro)                     (*  fd.h         *)
*/
/* ######################################################### */
/*                      InitCovData                          */
/* ######################################################### */
/*
   THIS FUNCTION PERFORMS A FEW INITIALIZATIONS THAT ARE NEEDED
   IN THE coverDataType VARIABLE BEFORE THE SWEEP STARTS.
*/
void InitCovData (cmdPtr, rDptr, covDptr, nrmInfPtr) 
      commandType    * cmdPtr ;
      ramDataRepType * rDptr ;
      coverDataType  * covDptr  ;
      normInfoType   * nrmInfPtr ;
{
   (covDptr->maxCoord) = (nrmInfPtr->maxCoord) ;
   if (!(cmdPtr->mincover))
   {
     (covDptr->numCellSizes) = (uLong) (rDptr->ulongBitNum)+1 ;
       /* ACTUALLY, THIS IS NOT USED WHEN MINCOVER IS FALSE. */
     (covDptr->magnifier) = 2.0 ;
   }
   else
   {
     (covDptr->numCellSizes) = (cmdPtr->numCellSizes) ;
     (covDptr->magnifier) = (cmdPtr->magnifier) ;
   }
   (covDptr->embedDim) = (rDptr->embedDim)  ;
   (covDptr->log2) = log(2.0) ;
   return ;
}
/* ######################################################### */
/*                        GetStatRAM                         */
/* ######################################################### */
/*
     IN THIS PROCEDURE, WE ALLOCATE STORAGE FOR THE STATISTICS
     THAT WILL BE GATHERED ON THE COVERING CELLS OF THE DATA
     SET.  THE TOTAL AMOUNT OF STORAGE ALLOCATE HERE IS
     O(numCellSizes + embeddingDimension) -- USUALLY NOT VERY
     MUCH.

     WE DID NOT HAVE TO MODIFY THIS FUNCTION TO GET THE
     mincover OPTION WORKING.  IT ALSO DOES NOT APPEAR THAT
     MANY CHANGES WILL BE NECESSARY WHEN AND IF THE mincover
     OPTION IS UPGRADED TO HANDLE INPUTS WITH EMBEDDING
     DIMENSION GREATER THAN 1.
*/
void  GetStatRAM (cdP)
      coverDataType  * cdP  ;
{  
   char s[11];
   size_t  ulS = sizeof(uLong), dS = sizeof(double) ;
   size_t  nCS = (size_t) (cdP->numCellSizes) ;
   size_t  eD = (size_t) (cdP->embedDim) ;
     /*
         WE KEEP TRACK OF THE NAME OF THE CALLING FUNCTION SO
         THAT gmalloc WILL BE ABLE TO PRINT THE NAME OF THE
         CALLER IN ANY ERROR MESSAGES.
     */
   strcpy(s, "GetStatRAM") ;
   
   (cdP->cellCount)      =(uLong *) gmalloc(s, "cellCount",      nCS * ulS);
   (cdP->pointCount)     =(uLong *) gmalloc(s, "pointCount",     nCS * ulS);
   (cdP->ptDiff)         =(uLong *) gmalloc(s, "ptDiff",          eD * ulS);
   (cdP->logCellSize)    =(double *)gmalloc(s, "logCellSize",     nCS * dS); 
   (cdP->negLogCellCount)=(double *)gmalloc(s, "negLogCellCount", nCS * dS);
   (cdP->information)    =(double *)gmalloc(s, "information",     nCS * dS);
   (cdP->sumSqrFreq)     =(double *)gmalloc(s, "sumSqrFreq",      nCS * dS);
   (cdP->logSumSqrFreq)  =(double *)gmalloc(s, "logSumSqrFreq",   nCS * dS);
   return ;
}
/* ######################################################### */
/*                      InitCellSizes                        */
/* ######################################################### */
/*
    THIS FUNCTION CREATES AN ARRAY CONTAINING THE LOGS (BASE
    TWO) OF THE CELL SIZES TO BE USED.  THIS HAS TO BE DONE
    DIFFERENTLY DEPENDING ON WHETHER THE -m (mincover) OPTION
    IS TO BE USED.
*/
void  InitCellSizes (cmdPtr, covDptr, nrmInfPtr) 
      commandType    * cmdPtr ;
      coverDataType  * covDptr  ;
      normInfoType   * nrmInfPtr ;
{
  uLong i ;
  
  if ( ! (cmdPtr->mincover) )
     for (i = 0; i < (covDptr->numCellSizes); i++)
           (covDptr->logCellSize)[i] = (double) i ;
  else
  {
    double size = (double) (nrmInfPtr->maxCoord) ;
    double mag = (covDptr->magnifier) ;
    double ln2 = (covDptr->log2) ;

    i = (covDptr->numCellSizes) ;

    do
    {
      i-- ;
      (covDptr->logCellSize)[i] = log(size)/ln2 ;
      size = size/mag ;
    }
    while (i > 0) ;
  }
}
/* ######################################################### */
/*                     PrintInitCovData                      */
/* ######################################################### */
/*

   THIS FUNCTION IS CALLED WHEN CHECKING OR DEBUGGING.

   "InitCovData" AND "InitCellSizes" PLACE CERTAIN VALUES INTO
   THE COVER DATA STRUCTURE before the SWEEP STARTS.  THIS
   FUNCTION HELPS US SEE IF THE VALUES ARE CORRECT.
   
*/
void PrintInitCovData (covDptr)
      coverDataType  * covDptr  ;
{
  uLong i ;
  double logx, x , ln2 = log(2.0);

  printf("\n" ) ;
  printf("The INITIAL COVER DATA is: \n") ;
  printf("\n" ) ;
  printf ("log2         = %f\n",  covDptr->log2 ) ;
  printf ("numCellSizes = %lu\n",  covDptr->numCellSizes) ;
  printf ("magnifier    = %f\n",  covDptr->magnifier) ;
  printf ("maxCoord     = %lu\n", covDptr->maxCoord) ;
  printf ("embedDim     = %lu\n", covDptr->embedDim ) ;
  printf("\n" ) ;
  printf ("index#            logCellSize            CellSize\n") ;
  printf("\n" ) ;
  
  i = (covDptr->numCellSizes) ;
  do
  {
    i-- ;
    logx = covDptr->logCellSize[i] ;
    x = exp(logx * ln2) ;
    printf ("%3lu%22.2f%24.2f\n",i,logx, x) ;
  } while (i > 0) ;
   printf ("\n") ;
   fflush(stdout);
   return ;
}
/* ######################################################### */
/*                        SweepData                          */
/* ######################################################### */
/*
    GET THE CELL COUNTS AND OTHER STATISTICS ASSOCIATED WITH
    THE COVERINGS.
*/
void  SweepData (cmdPtr, ramDataPtr, covDptr) 
      commandType    * cmdPtr ;
      ramDataRepType * ramDataPtr ;
      coverDataType  * covDptr  ;
{
  if (!(cmdPtr->mincover))
  {
    if    ((cmdPtr->debugging) && (cmdPtr->verbose))
          DBSweepData (ramDataPtr, covDptr) ;
    else  StdDataSweep (ramDataPtr, covDptr) ;
  }
  else mcovSweep(ramDataPtr, covDptr) ;
  return ;
}
/* ######################################################### */
/*                       StdDataSweep                        */
/* ######################################################### */
/*
    FOR THE STANDARD LIEBOVITCH-TOTH ALGORITHM, GET THE CELL
    COUNTS AND OTHER STATISTICS ASSOCIATED WITH THE COVERINGS.
*/
void  StdDataSweep (rDptr, covDptr) 
      ramDataRepType * rDptr ;
      coverDataType  * covDptr  ;
{
   uLong   prevIdx, curIdx, qSize, pairNum ;

   prevIdx = FrontIndex( &((rDptr->Q)[0]) ) ;
   InitStatRAM (covDptr) ;
   qSize = SizeQ ( &((rDptr->Q)[0]) ) ;
   for (pairNum = 1; pairNum < qSize; pairNum++)
   {
      curIdx = (rDptr->nextAfter)[prevIdx] ;
      {SEEKbitCHANGE(prevIdx, curIdx, rDptr, covDptr);}
      if ( covDptr->bitChFnd ) {UPDATEcovSTATS(rDptr, covDptr);}
      prevIdx = curIdx ;
   }
   FinStats (rDptr, covDptr) ;   /* FINISH-UP */
   return ;
}
/* ######################################################### */
/*                       InitStatRAM                         */
/* ######################################################### */
/*
     NECESSARY INITIALIZATIONS OF THE COVERING STATISTICS.
*/
void  InitStatRAM (covDptr)
      coverDataType  * covDptr  ;
{
   uLong curBitPos ;

   for(curBitPos = 0; curBitPos < (covDptr->numCellSizes); curBitPos++)
   { 
      (covDptr->cellCount)[curBitPos]    = 1; 
      (covDptr->sumSqrFreq)[curBitPos]   = 0.0;
      (covDptr->pointCount)[curBitPos]   = 1; 
      (covDptr->information) [curBitPos] = 0.0 ; 
   }
   return ;
}
/* ######################################################### */
/*                       FinStats                            */
/* ######################################################### */
/*
     WE ADD THE CONTRIBUTION DUE TO THE COUNTS WHICH REMAIN
     AFTER THE LAST DATA POINT HAS BEEN PROCESSED.  WE RENORMALIZE
     WITH cellCount[0], AND MASSAGE THE RAW DATA FROM THE SWEEP
     SO THAT IT IS IS READY FOR THE LEAST SQUARES FITTING.

     AFTER THIS PROCEDURE RUNS, FOR EACH BP FROM 0 TO
     numCellSizes, (covDptr->sumSqrFreq)[bP] CONTAINS THE SUM
     OF THE SQUARES OF THE SAMPLE PROBABILITIES OF POINTS IN
     ALL OCCUPIED CELLS WITH SIDE = 2^^bP.  ALSO,
     (covDptr->information)[bP] CONTAINS THE INFORMATION SUM
     FOR ALL THE OCCUPIED CELLS OF SIZE = 2^^bP.
*/
void FinStats (ramDptr, covDptr)
       ramDataRepType * ramDptr ;
       coverDataType  * covDptr  ;
{
  uLong bP ;
  double sProb ;
  
  for (bP = 0; bP < (covDptr->numCellSizes); bP++)
  {
    FinCapStats (covDptr, bP) ;
    sProb = (covDptr->pointCount)[bP]/(double)(ramDptr->dataLines) ;
    FinInfStats (ramDptr, covDptr, sProb, bP) ;
    FinCorrStats (ramDptr, covDptr, sProb, bP) ;
  }
  return ;
}
/* ######################################################### */
/*                      FinCapStats                          */
/* ######################################################### */
/*
    THIS PROCEDURE INPUTS A CELL COUNT AND OUTPUTS THE
    CORRESPONDING NEGATIVE LOG (BASE 2) CELL COUNT.  THE
    NEG-LOG-CELL-COUNTS ARE USED IN THE LEAST SQUARE FITTING
    THAT RESULTS IN THE CAPACITY DIMENSION ESTIMATE.
*/
void  FinCapStats  (covDptr, bitPos)
      coverDataType  * covDptr  ;
      uLong            bitPos ;
{
    double dTemp ;

    dTemp = (double) ((covDptr->cellCount)[bitPos]) ;
    dTemp = -log(dTemp)/(covDptr->log2) ;
    (covDptr->negLogCellCount)[bitPos] = dTemp ;

    return ;
}
/* ######################################################### */
/*                      FinInfStats                          */
/* ######################################################### */
/*
    THIS PROCEDURE PERFORMS THE FINAL INCREMENTING OF EACH
    INFORMATION STATISTIC AND THEN RE-NORMALIZES THE
    STATISTIC.

    THE RENORMALIZATION IS NEEDED BECAUSE DURING A STANDARD
    SWEEP, THE ACTUAL TOTAL NUMBER OF DISTINGUISHABLE DATA
    POINTS IS UNKNOWN, AND SO THE NUMBER OF LINES OF DATA IS
    USED INSTEAD IN THE CALCULATION OF SAMPLE PROBABILITIES.
*/
void  FinInfStats (ramDptr, covDptr, sampProb, bitPos)
        ramDataRepType * ramDptr  ;
        coverDataType  * covDptr  ;
        double           sampProb ;
        uLong            bitPos   ;
{
   double dTemp ;
      /*
         DO THE FINAL INCREMENT
      */
   dTemp = sampProb * log(sampProb)/(covDptr->log2) ;
   (covDptr->information)[bitPos] += dTemp ;
      /*
          RENORMALIZE
      */
   dTemp = (ramDptr->dataLines)/((double) ((covDptr->cellCount)[0])) ;
   (covDptr->information)[bitPos] *=  dTemp ;
   dTemp = (double) (ramDptr->dataLines) ;
   (covDptr->information)[bitPos] += log(dTemp)/(covDptr->log2) ;
   dTemp = (double)((covDptr->cellCount)[0]) ;
   (covDptr->information)[bitPos] -= log(dTemp)/(covDptr->log2);

   return ;
}
/* ######################################################### */
/*                     FinCorrStats                          */
/* ######################################################### */
/*
    THIS PROCEDURE PERFORMS THE FINAL INCREMENTING OF EACH
    SUM-SQUARE-FREQUENCY STATISTIC, THEN RE-NORMALIZES THE
    STATISTIC, AND THEN CALCULATES THE LOG (BASE 2) OF THE
    STATISTIC.

    THE RENORMALIZATION IS NEEDED BECAUSE DURING A STANDARD
    SWEEP, THE ACTUAL TOTAL NUMBER OF DISTINGUISHABLE DATA
    POINTS IS UNKNOWN, AND SO THE NUMBER OF LINES OF DATA IS
    USED INSTEAD IN THE CALCULATION OF SAMPLE PROBABILITIES.

    THE ESTIMATION OF THE CORRELATION DIMENSION REQURES THE LOG
    OF THE SUM-SQUARE-FREQUENCY STATISTICS.
*/
void  FinCorrStats (ramDptr, covDptr, sampProb, bitPos)
        ramDataRepType * ramDptr  ;
        coverDataType  * covDptr  ;
        double           sampProb ;
        uLong            bitPos   ;
{
    double dTemp ;
      /*
         DO THE FINAL INCREMENT
      */
    (covDptr->sumSqrFreq)[bitPos] += (sampProb * sampProb) ;
      /*
          RENORMALIZE
      */
    dTemp = (double) ((covDptr->cellCount)[0]) ;
    (covDptr->sumSqrFreq)[bitPos] *= (ramDptr->dataLines)/ dTemp ;
    (covDptr->sumSqrFreq)[bitPos] *= (ramDptr->dataLines)/ dTemp ;
      /*
          COMPUTE THE LOG OF THE STATISTIC.
      */
    (covDptr->logSumSqrFreq)[bitPos] = log((covDptr->sumSqrFreq)[bitPos]) ;
    (covDptr->logSumSqrFreq)[bitPos] /= (covDptr->log2) ;

    return ;
}
/* ######################################################### */
/*                         FindRange                         */
/* ######################################################### */
/*
    IN THE MAJORITY OF CASES, FindRange MERELY MAKES A CALL TO
    "RangeSweep".

    IN FACT, RANGE-CHECKING AND CONSISTENCY CHECKING IN
    getcom.c IS LIKELY TO INSURE THAT THE ERRORS TRAPPED HERE
    NEVER ACTUALLY GET PROPAGATED THIS FAR IN THE CALL CHAIN.
    HOWEVER IT'S NICE TO TRAP THESE ERRORS HERE AS A
    PRECAUTION.  THIS WAY, IF SOME CHANGE TO THE getcom.c
    MODULE "BREAKS" THE CHECKING DONE BY getcom.c, FindRange
    WILL NOT START PRODUCING INVALID RANGES.  THIS PROTECTS THE
    ROUTINES THAT DEPEND ON the OUTPUT of FindRange.  ALSO THE
    APPEARANCE OF THE ERROR MESSAGES WILL TIP OFF THE
    CODE-DEVELOPER THAT A PROBLEM HAS ARISEN IN getcom.c.

    WE HAVE GRAFTED THE -f OPTION (AND -m OPTION TOO) ONTO FD3
    IN A SOMEWHAT AD-HOC MANNER.  ONE UNDESIRABLE CONSEQUENCE
    IS THAT FindRange GETS CALLED UNNECESSARILY WHEN THE USER
    HAS SPECIFIED THE -f OPTION.  IN *THIS* CASE:

    1. THE VALUE OF covDptr->numCellSizes IS ulongBitNum+1,
       WHICH IS *NOT* THE NUMBER OF CELL SIZES THAT THE
       ESTIMATION PROCESS WILL EVENTUALLY BE BASED UPON,

    2. IT DOES NOT MAKE SENSE TO CHECK THE VALUE OF THE FINE
       MESH POINT AGAINST THE "BOGUS" VALUE OF covDptr->numCellSizes,

    3. THE RESULTS OF FindRange WILL NOT BE USED ANYWAY, AND 

    4. FindRefRange (OF coversf.c) WILL BE RESPONSIBLE FOR
       SETTING THE CUTOFF STAR BAR POSITIONS THAT ARE USED TO
       MAKE THE ESTIMATES.

    BELOW WE CHECK TO SEE IF THE user has specified both the -f
    and -p OPTIONs on the command line.  IF SO, WE JUST ALLOW
    FindRange TO TRANSFER THE SATURATION POINT AND FINE MESH
    POINT VALUES FROM THE COMMAND STRUCTURE INTO THE COVER DATA
    STRUCTURE.  FD3 WILL NOT USE THOSE VALUES ANYWAY.  FE3 WILL
    USE THE CORRESPONDING VALUES IN THE dimEstimDataType
    VARIABLE.

    ON THE OTHER HAND, IF THE USER HAS SET THE -p OPTION, BUT
    NOT THE -f OPTION, THEN THE VALUE OF covDptr->numCellSizes
    IS VALID AND THERE IS REASON TO PERFORM CHECKS THAT PREVENT
    AN OUT-OF-RANGE SET OF BOUNDS FROM BEING PRODUCED BY
    FindRange.  IN THIS CASE WE GO AHEAD AND MAKE THE
    APPROPRIATE CHECKS, AND EXIT THE PROGRAM IF THERE IS A
    PROBLEM.
*/
void  FindRange (covDptr, cP)
      coverDataType   * covDptr  ;
      commandType     * cP ;
{     
   if ( (covDptr->numCellSizes) < 2 )
   {
     printf("\n") ;
     printf("PROBLEM: \n") ;
     printf("FindRange was passed a value of numCellSizes < 2.\n") ;
     printf("At least 2 cell sizes are required for fitting\n") ;
     printf("lines whose slopes determine dimension estimates. \n") ;
     printf("\n") ;
     fflush(stdout) ;
     exit(EXIT_FAILURE) ;
   }
   if ( (cP->placeBars)) /* USER HAS REQUESTED A CELL RANGE. */
   {
      if (
           (cP->fine)
             || (
                  ((cP->saturationPoint) < (cP->fineMeshPoint))
                       && ((cP->fineMeshPoint) < (covDptr->numCellSizes))
                 )
         )
      {
        (covDptr->saturationPoint) = (cP->saturationPoint) ;
        (covDptr->fineMeshPoint)   = (cP->fineMeshPoint) ;
      }
      else
      {
        printf("\n") ;
        printf("PROBLEM in function FindRange: \n") ;
        printf("The cell-size bounds that you requested with\n") ;
        printf("the -p option are out of range.\n") ;
        printf("\n") ;
        fflush(stdout) ;
        exit(EXIT_FAILURE) ;
      }
   }
   else RangeSweep(covDptr, cP) ; 

   if (
        ((cP->debugging) || (cP->checking))
             && !(cP->fine)
      )
      ShowRange(covDptr) ;
   return ;
}
/* ######################################################### */
/*                        RangeSweep                         */
/* ######################################################### */
/*
    FD3 EMPLOYS RangeSweep UNLESS:

        A. THE fine (-f) OPTION IS USED,
        B. THE NUMBER OF CELL SIZES IS EXTREMELY SMALL, OR
        C. THE placeBars (-p) OPTION IS USED.
           
    RangeSweep PERFORMS A SEQUENTIAL SEARCH FOR THE PROPER
    VALUE OF THE saturationPoint.  IT ALSO SETS THE
    fineMeshPoint TO WHAT WE HOPE IS A REASONABLE VALUE.

    WE HAVE TO DETERMINE A RANGE OF CELL SIZES TO USE FOR THE
    DIMENSION ESTIMATES.  MANY CELL SIZES WILL BE "SATURATED"
    -- MUCH TOO SMALL COMPARED TO THE RESOLUTION OF THE DATA
    SET.  SOME CELL SIZES WILL BE TOO LARGE TO CAPTURE ANY
    USEFUL INFORMATION -- FOR EXAMPLE THE BOUNDING CELL
    ITSELF.  BELOW WE USE A SIMPLE HEURISTIC TO CALCULATE THE
    UPPER AND LOWER BOUNDS OF THE CELL SIZES TO BE USED.  OTHER
    HEURISTICS CAN BE APPLIED LATER.
      
    FOR THE saturationPoint, THE INDEX OF THE SMALLEST CELL
    SIZE USED IN THE ESTIMATES, WE COMPUTE THE LARGEST
    CELL-SIZE INDEX S SUCH THAT

             cellCount(S) > cellCount(1)/cutOffFAC

    "cellCount(T)" DENOTES THE NUMBER OF COMPUTED COVERING
    CELLS OF SIZE T, AND

             cutOffFAC =  maximum { (2^^(embedDim), 3 }

    WHERE embedDim IS THE EMBEDDING DIMENSION (THE NUMBER OF
    COORDINATES EACH POINT HAS), AND A^^B DENOTES
    EXPONENTIATION.  SO FOR EXAMPLE, FOR SMALL EMBEDDING
    DIMENSIONS, WE HAVE

             embedDim       cutOffFAC
            ---------     -------------
                1               5
                2               5
                3               8
                4              16

    (IN CERTAIN EXTREME CASES WHEN THE mincover (-m) IS USED,
    IT IS POSSIBLE THAT THERE BE *NO* INDEX S THAT SATISFIES THE
    INEQUALITY ABOVE.  IN THOSE CASES, THE saturationPoint IS
    SET TO THE SMALLEST INDEX: 0)

    COUNTS AT INDEXES LESS THAN saturationPoint WILL NOT BE
    USED IN THE ESTIMATE OF FRACTAL DIMENSION -- DISTORTION DUE
    TO SATURATION IS THE CONCERN.  

    UNLESS THE USER HAS SET THE NUMBER OF CELL SIZES WITH THE
    -m OPTION, FOR ALL PRACTICAL PURPOSES, cellCount(0) IS THE
    NUMBER OF DISTINCT POINTS IN THE DATA SET, BECAUSE CELLS OF
    SIDE = 1 ARE VERY SMALL COMPARED TO THE SIZE OF THE BIGGEST
    CELL (THE SIDE OF THE BIGGEST CELL IS ABOUT 4 BILLION IF AN
    UNSIGNED LONG INT IS IMPLEMENTED WITH 32 BITS ON THE
    PLATFORM COMPILER.  THE DATA POINTS ARE SCALED BY THE
    PROGRAM SO THAT THE DATA SET IS TOO "LARGE" TO FIT IN THE
    NEXT SMALLEST CELL SIZE.  THUS "1" IS THE SMALLEST
    DIFFERENCE IN VALUE THAT CAN BE RESOLVED.)  ONE CELL OF
    SIDE = 1, IN EFFECT, COVERS ONLY A SINGLE POINT OF THE DATA
    SET BECAUSE THE PROGRAM CAN'T RESOLVE POINTS WITH A SMALLER
    DIFFERENCE.
*/
void  RangeSweep (covDptr, cmdPtr)
      coverDataType   *  covDptr  ;
      commandType     * cmdPtr ;
{
  uLong i ;
  double log3 = log(3.0)/log(2.0), logCutFac ; 
  double logTotPoints, dtemp ;

  if   ( (covDptr->embedDim) > log3) logCutFac = (covDptr->embedDim) ;
  else logCutFac = log3 ;

  dtemp = (double) (covDptr->totDistPts) ;
  if   (cmdPtr->mincover) logTotPoints = log(dtemp)/log(2.0) ;
  else logTotPoints = -(covDptr->negLogCellCount)[0] ;
    
  i = 0 ;
  while ( i+1 < (covDptr->numCellSizes)
            && -((covDptr->negLogCellCount)[i+1])
                  > (logTotPoints - logCutFac) ) { i++ ; }

  (covDptr->saturationPoint) = i ;
       /*
           TRY TO PUT THE fineMeshPoint BEHIND THE LARGEST
           TWO CELL SIZES
       */
  if ( (covDptr->numCellSizes) > 5)
       (covDptr->fineMeshPoint) = (covDptr->numCellSizes) - 3 ;
                /*
                    IF THERE ARE VERY FEW CELL SIZES, PUSH THE
                    CUTOFF FORWARD BY ONE INDEX.
                */
  else if   ( (covDptr->numCellSizes) > 3)
            (covDptr->fineMeshPoint) = (covDptr->numCellSizes) - 2;
                 /*
                     IF THE NUMBER OF CELL SIZES IS REALLY TINY,
                     THEN LET THE fineMeshPoint BE THE LAST CELL
                     SIZE INDEX.
                 */
       else (covDptr->fineMeshPoint) = (covDptr->numCellSizes)- 1 ;
          
  return ;
}
/* ######################################################### */
/*                        ShowRange                          */
/* ######################################################### */
/*
     A PROCEDURE TO SHOW WHAT RANGE OF CELL INDEXES WAS CHOSEN
     FOR DOING THE LEAST-SQUARES ESTIMATES.
*/
void ShowRange (covDptr)
       coverDataType   *  covDptr  ;
{
   printf("\n") ;
   printf("The saturationPoint is: ") ;
   printf("%lu\n", covDptr->saturationPoint) ;
   printf("The fineMeshPoint is: ") ;
   printf("%lu\n", covDptr->fineMeshPoint) ;
   return ;
}
/* ######################################################### */
/*                     ReportCovData                         */
/* ######################################################### */
/*
      PRINT OUT THE TABLE CONTAINING COVERING STATISTICS.
*/
void ReportCovData (covDptr)
       coverDataType  * covDptr  ;
{
   CSTtitles      ();
   CSTbody        (covDptr);
   TwoPtEstTitles (covDptr) ;
   TwoPtEstBody   (covDptr) ;

   return ;
}
/* ######################################################### */
/*                          CSTtitles                        */
/* ######################################################### */
/*
     MAKE THE COLUMN TITLES FOR TABLE ONE.
*/
void CSTtitles ()
{
   printf("\n");
   printf("index");
   printf("  [log(eps)]"); printf("  [CellCnt]");
   printf("  [log(CellCnt)]");  printf("  [infrmtn]");
   printf("  [-log(SumSqrFreqs)]");   printf("\n") ;
   
   printf("       = [log(e)]"); printf("   = [N(e)] ");
   printf("   = [logN(e)]  ");  printf("  = [I(e)] ");
   printf("    = [-logSSF(e)]");   printf("\n\n");
  
  return ;
}
/* ######################################################### */
/*                          CSTbody                          */
/* ######################################################### */
/*
    PRINT OUT THE FIRST TABLE OF COVERING STATISTICS.
*/
void CSTbody (covDptr)
      coverDataType  * covDptr ;
{
  uLong j, satPt, finePt ;

  satPt = (covDptr->saturationPoint) ;
  finePt = (covDptr->fineMeshPoint) ;
  for(j = 0; j < (covDptr->numCellSizes); j++)
  {
     if ( (satPt < finePt) && (j == satPt ) ) MakeStarBar ();
     printf( "%5lu", j);
     printf( "%12.1f",    (covDptr->logCellSize)       [j]   );
     printf( "%11lu",     (covDptr->cellCount)         [j]   );
     printf( "%16.5f",   -((covDptr->negLogCellCount ) [j] ) );
     printf( "%11.5f",   -((covDptr->information     ) [j] ) );
     printf( "%21.5f\n", -((covDptr->logSumSqrFreq   ) [j] ) );
     if ( (satPt < finePt) && (j == finePt ) ) MakeStarBar ();
  }
  return ;
}
/* ######################################################### */
/*                      MakeStarBar                          */
/* ######################################################### */
void MakeStarBar ()
{
    printf ("**************************************");
    printf ("**************************************\n");

  return ;
}
/* ######################################################### */
/*                       TwoPtEstTitles                      */
/* ######################################################### */
/*
     MAKE THE COLUMN TITLES FOR TABLE TWO -- THE TABLE OF
     TWO-POINT ESTIMATES.
*/
void TwoPtEstTitles (covDptr) 
       coverDataType  * covDptr  ;
{
  printf ("\n\n\n");
  printf ("Two-Point Estimates of FD's ");
  printf ("based on magnifier M of %-.1f:",covDptr->magnifier);
  printf ("\n\n");
  printf("e             = cell size (side length, in fd3's scale)\n") ;
  printf("delta(e)      = log(M*e)    -  log(e)\n") ;
  printf("delta(log(N)) = log(N(e))   -  log(N(M*e))\n") ;
  printf("delta(I)      = I(e)        -  I(M*e)\n") ;
  printf("delta(logSSF) = logSSF(M*e) -  logSSF(e)\n") ;
  printf ("\n\n");
  printf("index      log(e)              ");
  printf("delta(log(N))   delta(I)        delta(logSSF)");
  printf("\n");
  printf("                               ");
  printf("-------------   --------        -------------") ;
  printf("\n");
  printf("                               ");
  printf("   delta(e)     delta(e)           delta(e)  ") ;
  printf("\n\n");
  return ;
}
/* ######################################################### */
/*                      TwoPtEstBody                         */
/* ######################################################### */
/*
   PRINT OUT THE TABLE OF TWO-POINT ESTIMATES.
*/
void TwoPtEstBody (c)
       coverDataType  * c ;
{
  uLong j, satPt, finePt ;
  double dTemp, lDiff ;

  satPt = (c->saturationPoint) ;
  finePt = (c->fineMeshPoint) ;
  for(j = 0; j < (c->numCellSizes)-1; j++)
  {
     if ( (satPt < finePt) && (j == satPt ) ) MakeStarBar ();

     lDiff = (c->logCellSize)[j+1] -((c->logCellSize)[j]) ;

     printf( "%5lu", j);
     printf( "%12.1f",(c->logCellSize)[j]);

     dTemp = ((c->negLogCellCount)[j+1]-(c->negLogCellCount)[j])/lDiff ;
     printf("%27.5f", dTemp);

     dTemp = ((c->information)[j+1]-(c->information)[j])/lDiff ;
     printf("%11.5f", dTemp);

     dTemp = ((c->logSumSqrFreq)[j+1]-(c->logSumSqrFreq)[j])/lDiff ;
     printf ("%21.5f", dTemp);

     printf ("\n") ;
     if ( (satPt < finePt) && (j+1 == finePt ) ) MakeStarBar ();
  }
  return ;
}
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */     
/*                       end covers.c                        */
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */
