
/* BEGIN COPYRIGHT NOTICE

Copyright (c) 1992 by John Sarraille and Peter DiFalco
(john@ishi.csustan.edu)

Copyright (c) 1998 by John Sarraille
(john@ishi.csustan.edu)

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that copyright notice and this permission
notice appear in supporting documentation.

The algorithm used in this program was inspired by the paper
entitled "A Fast Algorithm To Determine Fractal Dimensions By
Box Counting", which was written by Liebovitch and Toth, and
which appeared in the journal "Physics Letters A", volume 141,
pp 386-390, (1989).

This program is not warranted: use at your own risk.

END COPYRIGHT NOTICE */


/*
   FD3 PROGRAM FILE

Filename:  coversdb.c

OUTLINE OF CALLS:

  DBSweepData +
    FrontIndex
    InitStatRAM 
    SizeQ
    SeekBitChange +
      CheckBit +
        ISa  (macro)
    UpdateCovStats +
      ShowBitnSizes +
      ShowTemps +
      PrintUpdates +
    FinStats 
      FinCapStats 
      FinInfStats 
      FinCorrStats 
*/

#include "fd.h"

/* 
    THE FUNCTIONS IN THIS FILE: 
*/
void  DBSweepData     () ;
void  SeekBitChange   () ;
void  CheckBit        () ;         
void  UpdateCovStats  () ;
void  ShowBitnSizes   () ;
void  ShowTemps       () ;
void  PrintUpdates    () ;
/*   
      THE FUNCTIONS NOT IN THIS FILE,
      BUT CALLED FROM HERE:
*/
extern void   * gmalloc       () ;     /*  in mem.c     */
extern uLong    FrontIndex    () ;     /*  in queue.c   */
extern void     InitStatRAM   () ;     /*  in covers.c  */
extern uLong    SizeQ         () ;     /*  in queue.c   */
extern void     FinStats      () ;     /*  in covers.c  */
extern void     FinCapStats   () ;     /*  in covers.c  */
extern void     FinInfStats   () ;     /*  in covers.c  */
extern void     FinCorrStats  () ;     /*  in covers.c  */

/*
       MACROS CALLED FROM THIS FILE

    ISa  (macro)                     (*  fd.h         *)
*/
/* ######################################################### */
/*                       DBSweepData                         */
/* ######################################################### */
/*
    THIS IS A VERSION OF StdDataSweep TO BE USED WHEN DE-BUGGING
*/
void  DBSweepData (rDptr, covDptr) 
      ramDataRepType * rDptr ;
      coverDataType  * covDptr  ;
{  uLong   prevIdx, curIdx, qSize, pairNum ;

   prevIdx = FrontIndex ( &((rDptr->Q)[0]) ) ;
   InitStatRAM (covDptr) ;
   qSize = SizeQ ( &((rDptr->Q)[0]) ) ;
   for (pairNum = 1; pairNum < qSize; pairNum++)
   {
      curIdx = (rDptr->nextAfter)[prevIdx] ;
      SeekBitChange(prevIdx, curIdx, rDptr, covDptr);
      printf ("prevIdx: %lu, ", prevIdx) ; 
      printf ("curIdx: %lu, ", curIdx) ;
      printf ("(covDptr->bitPos): %lu\n", (covDptr->bitPos)) ; 
      printf ("\n") ;
      if ( covDptr->bitChFnd ) UpdateCovStats(rDptr, covDptr) ;
      prevIdx = curIdx ;
   }
   FinStats (rDptr, covDptr) ;   /* FINISH-UP */
   return ;
}
/* ######################################################### */
/*                      SeekBitChange                        */
/* ######################################################### */
/*
    SEE IF TWO DATA POINTS DIFFER, AND IF SO, DETERMINE THE MOST
    SIGNIFICANT BIT IN WHICH TWO RESPECTIVE COORDINATES DIFFER.
    IF NO DIFFERENCE IS FOUND, THEN (cDp->bitChFnd) = FALSE IS A
    POST-CONDITION.  IF A DIFFERENCE IS FOUND, THEN
    POST-CONDTIONS ARE THAT (cDp->bitChFnd) = TRUE AND
    (cDp->MSBdiff) = THE MOST SIGNIFICANT BIT POSITION IN WHICH
    TWO RESPECTIVE COORDINATES DIFFER.
*/
void SeekBitChange(pIdx, cIdx, rDp, cDp)
       uLong              pIdx, cIdx ;
       ramDataRepType  *  rDp ;
       coverDataType   *  cDp  ; 
{
   for ((cDp->coord) = 0; (cDp->coord) < (rDp->embedDim); (cDp->coord)++)
       (cDp->ptDiff)[(cDp->coord)] =   (rDp->pointList)[(cDp->coord)][pIdx]  
                                     ^ (rDp->pointList)[(cDp->coord)][cIdx] ;
   (cDp->bitPos)  =  (rDp->ulongBitNum) ;
   (cDp->bitChFnd)   =  FALSE ;
   while ( (!(cDp->bitChFnd)) && ((cDp->bitPos) > 0) )
   {
      (cDp->bitPos)-- ;
      CheckBit(cDp) ;
   }
   (cDp->MSBdiff) = (int) (cDp->bitPos) ;
   return ;
}
/* ######################################################### */
/*                           CheckBit                        */
/* ######################################################### */
/*
     CHECK BIT POSITION bitPos IN EACH COORDINATE OF THE
     POINT-DIFFERENCE ptDiff.  REPORT WHETHER A DIFFERENCE
     EXISTS BY USE OF THE bitChFnd FLAG.  
*/
void  CheckBit (cDp) 
       coverDataType   *  cDp  ; 
{
    (cDp->coord) = (cDp->embedDim) ;
    while ( ((cDp->coord) > 0) && (! (cDp->bitChFnd) ) ) 
    {
        (cDp->coord)-- ;
        (cDp->bitChFnd) = 
             ISa(((uLong) 1),(cDp->ptDiff)[(cDp->coord)],(cDp->bitPos)) ;
    }
   return ;
}
/* ######################################################### */
/*                      UpdateCovStats                       */
/* ######################################################### */
/*
     THE SWEEP IS THE CENTRAL PROCEDURE OF FD3.  THE UPDATES
     DONE DURING THE SWEEP ARE CODED IN THIS HELPER-FUNCTION.
     THE SO-CALLED "COVERING STATISTICS" ARE COMPUTED VIA
     INCREMENTAL CALLS TO THIS FUNCTION.
     
     EACH TIME THE SWEEP ENCOUNTERS A NEW DATA POINT, CERTAIN
     UPDATES MUST BE DONE TO THE COVERING STATISTICS.  THE MOST
     SIGNIFICANT BIT POSITION IN WHICH THE CURRENT POINT
     DIFFERS FROM THE PREVIOUS POINT (MSBdiff) DETERMINES WHICH
     STATISTICS MUST BE UPDATED, AND IN WHAT WAY.  WHEN MSBdiff
     = n, THE CURRENT POINT IS IN "NEW CELLS" RANGING IN SIZE
     FROM SIDE = 2^^n DOWN TO SIDE = 2^^0 = 1.  THE ORDERING
     GIVEN THE DATA POINTS BY THE PROGRAM'S (RADIX) SORT PHASE
     GUARANTEES THAT A "NEW CELL" IS REALLY "NEW" -- IT HAS NOT
     BEEN COUNTED PREVIOUSLY DURING THE SWEEP, WHEN SOME OTHER
     POINT IN THE DATA WAS THE CURRENT POINT.

     EXPLANATION OF EACH STEP:

     UPDATES CORRESPONDING TO ALL THE CELL SIZES HAVE TO BE
     DONE.  WE START WITH THE CELL SIZES BETWEEN 1 AND
     2^^MSBdiff.  THE MORE INTERESTING CHANGES ARE MADE IN THIS
     RANGE.
{
   for ((cDp->bitPos)=0; (cDp->bitPos)<=(cDp->MSBdiff); (cDp->bitPos)++)
   { 

     THE CURRENT POINT IS IN A *NEW* CELL WITH SIDE =
     2^^bitPos, SO WE INCREMENT THE CORRESPONDING COUNTER.

     ((cDp->cellCount)[(cDp->bitPos)])++ ;

     NOW CALCULATE A SAMPLE PROBABILITY THAT A RANDOM DATA
     POINT IS IN THE "OLD" CELL OF SIZE 2^^bitPos (THE "OLD"
     CELL IS THE ONE THAT CONTAINS THE PREVIOUS POINT.)  THIS
     PROBABILITY IS USED IN OUR ESTIMATES OF INFORMATION AND
     CORRELATION DIMENSIONS.  IN THE COURSE OF THE SWEEP, SUCH
     A PROBABILITY IS CALCULATED FOR EVERY OCCUPIED CELL --
     I.E. FOR EVERY CELL THAT CONTAINS ONE OR MORE OF THE DATA
     POINTS.  CONVERSELY, THE CALCULATION IS DONE *ONLY* FOR
     OCCUPIED CELLS.  WHEN DIVIDING BY (rDptr->dataLines)
     BELOW, WE ARE PROCEEDING AS IF THE NUMBER OF DATA LINES IN
     THE INPUT FILE WERE ALWAYS THE SAME AS THE NUMBER OF
     DISTINCT POINTS IN THE DATA SET.  THIS IS, OF COURSE, NOT
     TRUE, BECAUSE THERE COULD BE DUPLICATE POINTS.  AN
     ADJUSTMENT IS MADE FURTHER ON DURING THE COURSE OF THE
     PROCEDURE, WHEN A COUNT OF THE NUMBER OF "DISTINGUISHABLE"
     DATA POINTS IS AVAILABLE.  WE DIVIDE BY (rDptr->dataLines)
     NOW, MAINLY AS A WAY TO AVOID OVERFLOW OF THE double DATA
     TYPES BEING USED TO IMPLEMENT THE RUNNING SUMS sumSqrFreq
     AND information.  POSSIBLY MORE PRECISION COULD BE GAINED
     BY, FOR EXAMPLE, WAITING FOR sumSqrFreq TO GET CLOSE TO
     THE OVERFLOW VALUE, AND *THEN* DIVIDING BY THE SQUARE OF
     THE NUMBER OF DATA LINES.  (DOING THE EQUIVALENT THING FOR
     THE INFORMATION STATISTIC PRESENTS MORE OF A PROBLEM!)

     (cDp->sProb)=(cDp->pointCount)[(cDp->bitPos)]/(double)(rDp->dataLines);

     SINCE THE SWEEP HAS JUST FINISHED COUNTING ALL POINTS IN
     THE "OLD" CELL, WE MAY NOW COMPUTE WHAT THOSE POINTS
     CONTRIBUTE TO THE RUNNING SUMS BELOW.  AFTER THAT, WE ARE
     FREE TO RE-USE THE (cDp->pointCount) SLOT TO COUNT THE
     NUMBER OF POINTS IN THE CURRENT CELL.
                         
     (cDp->sumSqrFreq)[(cDp->bitPos)] += ((cDp->sProb) * (cDp->sProb)) ;
     (cDp->information)[(cDp->bitPos)] += ((cDp->sProb) * log((cDp->sProb))
                                             / (cDp->log2) ) ;

     WE INITIALIZE THE COUNTER TO 1, TO REFLECT THE ONE "NEW"
     (CURRENT) POINT IN THE "NEW" (CURRENT) CELL.

     (cDp->pointCount)[(cDp->bitPos)] = 1 ;
   }

     FOR CELLS OF SIDE = 2^^(MSBdiff+1) UP TO SIDE =
     2^^(numCellSizes-1), THE CURRENT POINT IS IN THE "OLD"
     CELLS -- NOT IN A "NEW" CELL.  SO FOR THESE CELL SIZES, WE
     MERELY INCREMENT THE POINT COUNTER TO REFLECT THE
     DISCOVERY OF THE CURRENT POINT.

   for (   (cDp->bitPos)=(cDp->MSBdiff)+1; 
           (cDp->bitPos)<(cDp->numCellSizes); 
           (cDp->bitPos)++)
       ( (cDp->pointCount)[(cDp->bitPos)] )++ ;
*/
void UpdateCovStats(rDp, cDp) 
       ramDataRepType  * rDp ;
       coverDataType   * cDp  ;
{
   ShowBitnSizes (cDp) ; 
   for ((cDp->bitPos)=0; (cDp->bitPos)<=(cDp->MSBdiff); (cDp->bitPos)++)
   { 
     ((cDp->cellCount)[(cDp->bitPos)])++ ;

     (cDp->sProb)=(cDp->pointCount)[(cDp->bitPos)]/(double)(rDp->dataLines);
     (cDp->sumSqrFreq)[(cDp->bitPos)] += ((cDp->sProb) * (cDp->sProb)) ;
     (cDp->information)[(cDp->bitPos)] += ((cDp->sProb) * log((cDp->sProb))
                                             / (cDp->log2) ) ;
     (cDp->pointCount)[(cDp->bitPos)] = 1 ;
     ShowTemps (cDp) ; 
   }
   for (   (cDp->bitPos)=(cDp->MSBdiff)+1; 
           (cDp->bitPos)<(cDp->numCellSizes);  (cDp->bitPos)++)
       ( (cDp->pointCount)[(cDp->bitPos)] )++ ;
   PrintUpdates (cDp) ;

   return ;
}
/* ######################################################### */
/*                    ShowBitnSizes                          */
/* ######################################################### */
/*
    THIS PROCEDURE IS TO BE USED FOR SOME CLOSE CHECKING OF
    WHAT THE PROGRAM IS DOING.  IT IS INTENDED TO BE USED ONLY
    IF WE ARE DEBUGGING *AND* WE HAVE ASKED FOR VERBOSE
    OUTPUT.  WHEN EMPLOYED, FOR EACH NEXT POINT IN THE DATA, IT
    WILL REPORT THE MOST SIGNIFICANT BIT POSITION IN WHICH IT
    DIFFERS FROM THE PREVIOUS POINT.
*/
void  ShowBitnSizes (cDp) 
       coverDataType   * cDp  ;
{  
    printf ("UpdateCovStats is running.\n") ;
    printf ("MSBdiff = %d; ", (cDp->MSBdiff) ) ;   
    printf ("numCellSizes = %lu", (cDp->numCellSizes)) ;  
    printf ("\n\n") ;   

    return ;
}
/* ######################################################### */
/*                   ShowTemps                          */
/* ######################################################### */
/*
    THIS IS A PROCEDURE THAT IS USED ONLY IN VERBOSE DEBUGGING
    MODE.  bitPos INDICATES THE CURRENT CELL SIZE AND sProb
    GIVES THE FREQUENCY OF DATA POINTS IN THE CURRENT CELL.
*/
void  ShowTemps (cDp) 
       coverDataType   * cDp  ;
{
    printf ("bitPos = %lu; ", (cDp->bitPos) ) ;
    printf ("sProb = %f\n", (cDp->sProb)) ;

    return ;
}
/* ######################################################### */
/*                       PrintUpdates                        */
/* ######################################################### */
/*
    THIS IS A PROCEDURE THAT IS USED ONLY IN VERBOSE DEBUGGING
    MODE.  IT PRINTS OUT THE VALUES OF THE STATISTICS BEING
    GATHERED BY THE SWEEP.  IT IS USED TO SHOW THESE VALUES
    AFTER EACH NEW POINT IS ENCOUNTERED. 
*/
void  PrintUpdates   (covDptr)
        coverDataType   * covDptr  ;
{
   uLong bitPos ;
   
   printf ("The COVER STAT UPDATES ARE: \n\n" ) ;
   printf ("bitPos  cellCount  pointCount  sumSqrFreq  information\n\n") ;
   for (bitPos = 0; bitPos < (covDptr->numCellSizes); bitPos++)
   {
      printf ("%6lu"  , bitPos ) ;
      printf ("%11lu", (covDptr->cellCount)   [bitPos] ) ;
      printf ("%12lu", (covDptr->pointCount)  [bitPos] ) ;
      printf ("%12f", (covDptr->sumSqrFreq)  [bitPos] ) ;
      printf ("%13f", (covDptr->information) [bitPos] ) ;
      printf ("\n") ;
   }
   printf ("\n\n") ;
   return ;
}
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */     
/*                      end coversdb.c                       */
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */
